#!/usr/bin/env/python
# Tells Linux systems how to run this file.
"""
Morningside
by Kris Schnee.

"""

##___________________________________
##          ~Header~
##___________________________________

__author__ = "Kris Schnee"
__version__ = "2009.9.25"
__license__ = "Public Domain"


##___________________________________
##      ~Imported Modules~
##___________________________________

## Standard
import os
import random
import time

## Third-party

## Mine


##___________________________________
##            ~Constants~
##___________________________________

GRAPHICS = False ## Set this to True if you want to try the demo.
if GRAPHICS:
    import pygame
    from pygame.locals import *
    screen = pygame.display.set_mode((400,400))

"""Terrain types.
Some cost energy to enter if the "energy" rule is in play.
Others are impassible or even kill any character that enters."""
TERRAINS = {"water":{"cost":3},
            "grass":{},
            "sand":{},
            "dirt":{},
            "block":{"passable":False},
            "forest":{"cost":2},
            "mountains":{"passable":False},
            "iron":{},
            "oak":{},
            "pit":{"passable":False},
            "lava":{"deadly":True},
            }
TERRAIN_ABBREVS = {}
for t in TERRAINS:
    TERRAINS[t].setdefault("cost",1)
    TERRAINS[t].setdefault("passable",True)
    TERRAINS[t].setdefault("abbrev",t[0])
    TERRAIN_ABBREVS[TERRAINS[t]["abbrev"]] = t

TIME_FORMAT = "%Y %b %d %H:%M:%S"
TIME_SHORTFORMAT = "%H:%M"

LEVEL_DIRECTORY = "levels"

##___________________________________
##            ~Classes~
##___________________________________

class Entity:
    def __init__(self,**options):
        self.name = options.get("name","Entity")
        self.color = options.get("color","grey")
        self.shape = options.get("shape","cube")
        self.size = options.get("size",1)
        """Coords here are for reading only. Changing them will not
        change the actual location, ie. which Tile holds the Entity."""
        self.coords = (None,None)
        self.alive = False
        self.held_item = None
        self.food_value = options.get("food value",0)
        self.is_door = options.get("is door",False)
        self.is_key = options.get("is key",False)
        self.untakeable = options.get("untakeable",False)
        if self.is_door:
            self.untakeable = True
        self.ID = options.get("ID")
        self.image = options.get("image","default") ## Optional graphics file name

    def SetCoords(self,coords=(None,None)):
        """Set internal coordinates. DOES NOT call Tile.Add."""
        self.coords = coords

    def Get(self,entity):
        """Put this in my 'held_item' slot."""
        if self.held_item:
            return False
        else:
            self.held_item = entity
            entity.coords = (None,None)
            return True

    def Drop(self):
        """Remove what I'm holding (if anything) and return it."""
        if not self.held_item:
            return False
        entity = self.held_item
        self.held_item = None
        return entity




class Character( Entity ):
    """An Entity that can perform actions."""
    def __init__(self,**options):
        Entity.__init__(self,**options)
        self.name = options.get("name","Character")
        self.shape = options.get("shape","humanoid")
        self.size = options.get("size",10)
        self.alive = True
        self.food_value = 0 ## !
        self.is_door = False
        self.is_key = False
        self.energy = options.get("energy",100)
        self.energy_max = 100
        self.vision_range = None ## N/A, not yet implemented
        self.command = None ## When all characters enter a command, turn starts.
        self.command_history = []
        self.sensed_events = [] ## Things I've sensed & should report.
        self.system_messages = []
        self.output_file_name = options.get("output_file_name","to_"+self.name.lower()+".txt")
        self.input_file_name = options.get("input_file_name","from_"+self.name.lower()+".txt")
        self.proctor = options.get("proctor",False) ## Doesn't lose energy.

    def Sense(self,sender,what):
        self.sensed_events.append(sender+": "+what)

    def SenseSystemMessage(self,what):
        """Record a special message."""
        self.system_messages.append(what)

    def ClearInput(self):
        self.sensed_events = []
        self.system_messages = []

    def SetCommand(self,command):
        """This is what I'll do this turn..."""
        self.command = command
        self.command_history.append(command)

    def ChangeEnergy(self,amount=0):
        """Gain/lose energy. Returns True if I'm still at >0 energy."""
        if amount > 0:
            self.energy = min(self.energy + amount, self.energy_max)
            return True
        else:
            if self.proctor:
                return True ## Never mind.
            self.energy -= amount
            return self.energy > 0

    def FillEnergy(self):
        self.energy = self.energy.max




class Tile:
    def __init__(self,coords,terrain="w"):
        self.coords = coords
        self.SetTerrain(terrain)
        self.contents = None ## Holds one entity.

    def SetTerrain(self,terrain):
        self.terrain = terrain
        self.passable = TERRAINS[terrain]["passable"]
        self.cost = TERRAINS[terrain].get("cost",0)
        self.deadly = TERRAINS[terrain].get("deadly",False)

    def Put(self,entity):
        """Put entity here if possible. Return T/F."""
        if self.contents:
            return False
        self.contents = entity
        entity.SetCoords(self.coords)
        return True

    def Take(self):
        """Remove entity from this tile. Return entity/None."""
        if self.contents:
            entity = self.contents
            entity.coords = (None,None)
            self.contents = None
            return entity
        return None




class World:
    """A simple worldsim for AI testing.

    Features:
    -Turn-based
    -Tile-based
    -Simple string input (send eg. "go n" to go north)
    -Entities with very simple properties to observe
    -"Player Characters" as subsets of entities, ie. they follow the same rules
    -PCs have simple physical needs (optional)
    -Simple interface using reading/writing of text files
    -Logging
    """
    def __init__(self,**options):
        self.level_name = options.get("level_name","Untitled Level")
        self.time = 0 ## Internal time, in rounds.
        self.seed = options.get("seed","")
        ## Create a grid of Tiles.
        self.size = options.get("size",(8,8))
        self.world = []
        self.CreateBlankGrid()

        ## Dictionary of all entities in the world by ID#.
        self.entities = {}

        ## Characters are listed here and stored in the world's Tiles.
        ## Note: Characters act in the order listed here (ie. order of joining).
        self.characters = []

        ## Definitions of recurring items, eg. if you want 10 identical rocks.
        self.object_types = {}

        ## Log
        ## Entries have "tags" and text.
        self.log = []
        self.echo_log = options.get("echo_log",True) ## Print when logging.
        self.log_comic = options.get("log_comic",False)
        self.comic_script = [[]] ## Mainly for fun -- assemble a log as comic!

        ## Entity IDs
        ## A guaranteed-unique-per-session ID number for each entity.
        ## In this version, AIs are told these IDs explicitly.
        ## That feature lets them know "the box here is the same one as before".
        self.highest_ID = 0

        ## Starting locations for characters
        ## Only as many characters can play as there are "spawn points".
        self.starting_points = [(0,0),
                                (self.size[0]-1,0),
                                (0,self.size[1]-1),
                                (self.size[0]-1,self.size[1]-1)
                                ]
        self.starting_points_available = [True,True,True,True]
        ## Can players pick one?
        self.starting_points_selectable = options.get("starting_points_selectable",False)

        ## Optional graphics system
        self.use_graphics = options.get("use_graphics")
        if self.use_graphics:
            import pygame
            self.tiles = {}
            self.tiles["grass"] = pygame.surface.Surface((50,50))
            self.tiles["grass"].fill((0,200,0))
            self.tiles["dirt"] = pygame.surface.Surface((50,50))
            self.tiles["dirt"].fill((100,50,50))
            self.tiles["forest"] = pygame.surface.Surface((50,50))
            self.tiles["forest"].fill((0,150,0))
            self.tiles["water"] = pygame.surface.Surface((50,50))
            self.tiles["water"].fill((0,50,200))
            self.tiles["sand"] = pygame.surface.Surface((50,50))
            self.tiles["sand"].fill((220,220,50))
            self.tiles["block"] = pygame.surface.Surface((50,50))
            self.tiles["block"].fill((250,250,250))
            self.tiles["mountains"] = pygame.surface.Surface((50,50))
            self.tiles["mountains"].fill((250,250,250))
            self.tiles["player"] = pygame.surface.Surface((50,50))
            pygame.draw.circle(self.tiles["player"],(255,0,255),(24,24),10)

    def MakeSeed(self):
        seed = ""
        values = "0123456789ABCDEF"
        random.seed()
        for n in range(5):
            seed += values[random.randint(0,15)]
        return seed

    def CreateBlankGrid(self):
        """Make a blank Tile list."""
        self.world = []
        for x in range(self.size[0]):
            self.world.append([])
            for y in range(self.size[1]):
                self.world[-1].append( Tile((x,y),"grass") )

    def SetRandomTerrain(self,seed=None):
        """Establish the terrain."""
        if seed:
            random.seed(seed)
        for x in range(self.size[0]):
            for y in range(self.size[1]):
                n = random.randint(0,99)
                if n < 20:
                    terrain = "water"
                elif n < 30:
                    terrain = "dirt"
                elif n < 40:
                    terrain = "sand"
                elif n < 43:
                    terrain = "block"
                elif n < 46:
                    terrain = "forest"
                elif n < 50:
                    terrain = "mountains"
                else:
                    terrain = "grass"
                self.world[x][y].SetTerrain(terrain)

    def PickRandomTile(self,terrain=None):
        """Pick a random tile containing nothing, with certain terrain."""
        t = []
        for x in range(self.size[0]):
            for y in range(self.size[1]):
                ok = (not self.world[x][y].contents) and ((not terrain) or self.world[x][y].terrain[0] == terrain)
                if ok:
                    t.append([x,y])
        n = random.randint(0,len(t)-1)
        return t[n]

    def LoadLevel(self,filename="default_level.txt"):
        """Set the terrain and characters based on a file.

        Each line in a level file should follow one of these conventions:
        # starts with "#"
        -- Comment. Is ignored.
        key:value
        -- sets one of the level's features such as the tile layout.
        *object_name:param1=spam,param2=bacon,param3=eggs
        -- Defines an object that will be put on the board.
        
        When adding objects, you can use the parameter "copy=True"
        to imitate an existing object. Eg. "*rock:coords=4,2/copy=True"
        will make a rock identical to a rock you've already described,
        but for any parameters you've specified (here, the coords).
        """

        if not filename.endswith(".txt"):
            filename += ".txt"
        self.Log("setup","Loading new level from \""+filename+"\".")
        ## First, clear any existing characters and rules.
        self.characters = []
        self.rules = {}
        try:
            f = open(os.path.join(LEVEL_DIRECTORY,filename),"r")
        except:
            self.Log("error,IO","Couldn't load level file \""+filename+"\".")
            raise SystemError("Couldn't load level file.")
        data = f.read()
        f.close()

        data = data.replace("_"," ")
        
        self.level_name = filename.split(".")[0].title().replace("_"," ")
        layout = None
        
        lines = data.split("\n")
        d = {}
        entities_to_make = []
        for line in [l for l in lines if l and l[0] != "#"]:
            kv = line.split(":")
            key = kv[0].rstrip(" ").lower().replace("_"," ")
            value = kv[1].lstrip(" ")
            if key.startswith("*"): ## Special case: Object description
                entities_to_make.append((key[1:],value))
            else:
                d[key] = value

        ## Seed value: Sets randomness generator. Don't use w/ a random layout.
        value = d.get("seed")
        if value:
            self.seed = value
        else:
            self.seed = self.MakeSeed()
        random.seed(self.seed)
        self.Log("setup","Seeded the RNG with this code: "+self.seed)

        value = d.get("starting points")
        if value:
            self.starting_points = []
            starting_points = value.split("/")
            for entry in starting_points:
                s = entry.split(",")
                self.starting_points.append((int(s[0]),int(s[1])))
            self.starting_points_available = [True]*len(self.starting_points)

        value = d.get("starting points available")
        if value:
            self.starting_points_selectable = value.lower() == "true"
        value = d.get("size")
        if value:
            s = tuple(value.split(","))
            self.size = (int(s[0]),int(s[1]))
        value = d.get("layout")
        if value:
            layout = value
        value = d.get("notes")
        if value:
            self.Log("setup","Notes: "+value)
        value = d.get("energy")
        if value:
            self.rules["energy"] = True
            self.rules["starting energy"] = int(d.setdefault("starting energy",100))
            self.rules["energy drain rate"] = int(d.setdefault("energy drain rate",1))
        else:
            self.rules["energy"] = False

        ## Now, turn the layout string into a set of Tile objects.
        if not d.has_key("layout"):
            self.Log("error,io","The level file didn't include a layout.")
            raise SystemError("No layout in level file.")
        layout = d["layout"]
        if layout.startswith("random"):
            ## Special: Set layout to random.
            seed = self.seed
            r = layout.split(" ")
            if (len(r) > 1) and r[1]:
                seed = r[1].lower()
                ## Use second word for a seed, if one is given.
                self.level_name = "* "+seed+" *"
            self.CreateBlankGrid()
            self.SetRandomTerrain(seed)
        else:
            try:
                assert len(layout) == self.size[0]*self.size[1]
            except:
                raise SystemError("Layout length doesn't match level's stated size. Should be a string with no spaces, x*y long.")
            terrain_types = TERRAIN_ABBREVS.keys()
            self.world = []
            for x in range(self.size[0]):
                self.world.append([])
                for y in range(self.size[1]):
                    terrain = layout[(y*self.size[0]) + x]
                    if not terrain in terrain_types:
                        self.Log("error,io","I don't recognize the terrain type '"+terrain+"'; allowed types are: "+str(terrain_types))
                    self.world[-1].append( Tile((x,y),TERRAIN_ABBREVS[terrain]) )

        ## Place entities on those Tiles.
        for e in entities_to_make:
            ## Format: "*name:color=?,shape=?,other_parameter=?"
            ## Example: "*acorn:,color=brown,shape=round"
            ## Note: Text has already had any "_" replaced with " ".
            params = {"name":e[0],"coords":"?"}
            keys_that_are_for_numbers = ["food value","size"]
            keys_that_are_for_bool = ["is key","is door"]
            p = e[1].replace(", ",",").replace(":","=").split("/")
            for item in p:
                i = item.split("=")
                params[i[0]] = i[1]
                if i[0] == "coords":
                    if i[1].startswith("?"):
                        ## Special: put it randomly on certain terrain.
                        if len(i[1])>1:
                            terrain = i[1][1] ## eg. "g" == "any open grass tile"
                        else:
                            terrain = "" ## Any open tile.
                        params["coords"] = self.PickRandomTile(terrain)
                    else:
                        coords = i[1].split(",")
                        params["coords"] = [int(coords[0]),int(coords[1])]
                elif i[0] in keys_that_are_for_numbers:
                    params[i[0]] = int(i[1])
                elif i[0] in keys_that_are_for_bool:
                    if i[1].lower() == "true":
                        params[i[0]] = True
                    else:
                        params[i[0]] = False

            if params.get("copy"):
                ## Make this object based on one already defined.
                try:
                    copy = self.object_types[params["name"]].copy()
                except:
                    raise SystemError("The level file requested a copy of an object called '"+params["name"]+", but that object wasn't already defined.")
                copy.update(params)
                params = copy
            else:
                ## Store this object's stats in case copies are requested.
                self.object_types[params["name"]] = params
            ## Now we actually build the object and add it to the world.
            entity = self.Create(**params)

        ## Set up some optional rules.
        if self.rules:
            self.Log("setup","Rules: "+str(self.rules))
        self.Log("setup","Finished loading level \""+self.level_name+"\".")

    def Log(self,tags,text=""):
        """Record an event."""
        self.log.append((time.strftime(TIME_SHORTFORMAT,time.localtime()),tags,str(text)))
        if self.echo_log:
            print "["+tags+"] "+str(text)
        if self.log_comic and "comic" in tags:
            self.comic_script[-1].append(text)

    def WriteLog(self,filename="log.txt"):
        if not filename.endswith(".txt"):
            filename += ".txt"
        f = open(filename,"w")
        f.write("------ Log file for worldsim \"Morningside\" ------\n")
        for entry in self.log:
            f.write(entry[0]+" ["+entry[1].title()+"] "+str(entry[2])+"\n")
        f.write("Log saved at time: "+time.strftime(TIME_FORMAT,time.localtime()))
        f.close()

    def Create(self,**options):
        """Create an Entity and place it in the world."""
        coords = options.get("coords",(0,0))
        if self.world[coords[0]][coords[1]].contents:
            self.Log("error","Can't create an Entity at coords "+str(coords)+"; already occupied.")
            return None
        options["ID"] = str(self.highest_ID).zfill(2) ## eg. "01"
        self.highest_ID += 1
        entity = Entity(**options)
        self.entities[entity.ID] = entity
        self.PlaceAt(entity,coords)
        self.Log("setup,object","Created object with params: "+str(options))
        return entity

    def CreateCharacter(self,**options):
        """Like Create.

        Starting locations are limited to a specific set.
        If a "selectable" flag is set, players can pick one of these.
        Otherwise, the first available one is chosen. Note: login order counts.
        
        Each point can only be used once, limiting the number of players.
        This isn't an FPS game with "spawn points" chosen for tactics;
        the point is to allow for predictable, repeatable board setups."""
        
        desired_coords = options.get("coords")
        coords = None ## Below we'll decide the actual coords used.

        if not True in self.starting_points_available:
            self.Log("error,boardfull","Can't add a new character; all start points used.")
            return
        
        if desired_coords and self.starting_points_selectable:
            for n in range(len(self.starting_points)):
                if self.starting_points_available[n] and self.starting_points[n] == coords:
                    self.starting_points_available[n] = False
                    coords = desired_coords
                    break
        if not coords:
            ## No desired coords, or coords unavailable, or selection disallowed
            ## Assign the first available starting point.
            for n in range(len(self.starting_points)):
                if self.starting_points_available[n]:
                    self.starting_points_available[n] = False
                    coords = self.starting_points[n]
                    break

        if self.world[coords[0]][coords[1]].contents:
            self.Log("error","Can't create a Character at coords "+str(coords)+"; already occupied.")
            return None

        if self.rules["energy"]:
            options["energy"] = self.rules["starting energy"]
        
        options["ID"] = str(self.highest_ID).zfill(2) ## eg. "01"
        self.highest_ID += 1
        character = Character(**options)
        self.entities[character.ID] = character
        self.PlaceAt(character,coords)
        self.characters.append(character)
        character.SenseSystemMessage("you: "+character.ID)
        character.SenseSystemMessage("level_size: "+str(self.size[0])+","+str(self.size[1]))
        character.SenseSystemMessage("rules: "+str(self.rules))
        self.Log("setup,login","Created a new character, \""+character.name+"\", at location: "+str(character.coords)+".")
        return character

    def PlaceAt(self,entity,coords):
        """Put this entity at these coords.

        Works whether or not the entity already has a location."""
        x,y = coords
        if entity.coords:
            ## Remove it from its current tile.
            self.world[x][y].Take()
        OK = self.world[x][y].Put(entity)
        return OK

    def Move(self,entity,direction):
        """Try to move in this direction (n/s/e/w). Return T/F."""
        ## Where is the entity trying to go?
        x,y = entity.coords[0], entity.coords[1]
        if direction == "n":
            y -= 1
        elif direction == "s":
            y += 1
        elif direction == "w":
            x -= 1
        elif direction == "e":
            x += 1
        else:
            ## Syntax error.
            self.Log("error,syntax","Syntax: 'go n/s/e/w'")
            return False
        ## Is the destination outside the world boundaries?
        if (x<0) or (y<0) or (x>=self.size[0]) or (y>=self.size[1]):
            self.Log("fail","Couldn't move "+direction+"; out of bounds.")
            return False
        ## Is the destination a passable type of terrain (ie. not a wall)?
        if not self.world[x][y].passable:
            self.Log("fail","Couldn't move "+direction+"; impassible.")
            return False
        ## Is the destination occupied?
        if self.world[x][y].contents:
            self.Log("fail","Oof! The destination was occupied.")
            return False
        ## It's OK to move. Do it!
        self.world[entity.coords[0]][entity.coords[1]].Take()
        OK = self.world[x][y].Put(entity)
        if not OK:
            self.Log("error,WTF","Unknown error.")
            raise SystemError("Movement somehow blocked in empty tile!")
        self.Log("success","Moved "+direction+".")

        ## Successful movement. Check hazards...
        terrain = self.world[x][y].terrain
        deadly = TERRAINS[terrain].get("deadly")
        if deadly:
            self.Log("fail,comic",entity.name+" walked into deadly "+terrain+" terrain!")
            self.Kill(entity)
        if self.rules.get("energy"):
            cost = TERRAINS[terrain].get("cost")
            survived = entity.ChangeEnergy(-cost)
            if not survived:
                self.Kill(entity)
        return True

    def Kill(self,character):
        """Destroy an entity.

        If the entity was alive, ie. a character, it can no longer play,
        but it stays on the board and is treated like an object.
        Non-living entities that're destroyed get removed from play."""
        if character.alive:
            self.Log("death,comic",character.name+" has died.")
            self.AllSense(character.ID,"died")
            character.energy = 0
            character.alive = False
        else:
            self.Log("error,syntax","The Kill function should only be used for living entities.")

    def Take(self,entity,direction):
        """Try to take a non-living entity in this direction. Return T/F."""
        ## Is the taker already carrying something?
        if entity.held_item:
            self.Log("fail","Already holding an item.")
            return False
        ## Where is the entity trying to take from?
        x,y = entity.coords[0], entity.coords[1]
        if direction == "n":
            y -= 1
        elif direction == "s":
            y += 1
        elif direction == "w":
            x -= 1
        elif direction == "e":
            x += 1
        else:
            ## Syntax error.
            self.Log("error,syntax","Syntax: 'take n/s/e/w'")
            return False
        ## Is the target outside the world boundaries?
        if (x<0) or (y<0) or (x>=self.size[0]) or (y>=self.size[1]):
            self.Log("fail","Couldn't take from "+direction+"; out of bounds.")
            return False
        ## Is the target tile empty?
        if not self.world[x][y].contents:
            self.Log("fail","Nothing to take there.")
            return False
        ## Is the entity living or otherwise untakeable?
        if self.world[x][y].contents.alive or self.world[x][y].contents.untakeable:
            self.Log("fail","You can't take this.")
            return False

        ## OK!
        target_entity = self.world[x][y].Take() ## Target's coords now == None.
        entity.Get(target_entity)
        self.Log("success","Took #"+target_entity.ID+ " "+target_entity.name+" from "+direction+".")
        return True

    def Drop(self,entity,direction):
        """Try to drop a held entity. Return T/F."""
        ## The entity must already be holding something.
        if not entity.held_item:
            self.Log("fail","Not holding an item.")
            return False
        ## Where is the entity trying to drop to?
        x,y = entity.coords[0], entity.coords[1]
        if direction == "n":
            y -= 1
        elif direction == "s":
            y += 1
        elif direction == "w":
            x -= 1
        elif direction == "e":
            x += 1
        else:
            ## Syntax error.
            self.Log("errorSYN","Syntax: 'drop n/s/e/w'")
            return False
        ## Is the target outside the world boundaries?
        if (x<0) or (y<0) or (x>=self.size[0]) or (y>=self.size[1]):
            self.Log("fail","Couldn't move "+direction+"; out of bounds.")
            return False

        ## Target tile must be passable.
        if not self.world[x][y].passable:
            self.Log("fail","Can't drop items onto impassible terrain.")
            return False

        ## Target tile must be empty. Exception: Door objects.
        if self.world[x][y].contents:
            item = self.world[x][y].contents
            print "Object is in target space: "+item.name

            ## Special case: Dropping keys on matching-color doors removes both.
            if item.is_door and entity.held_item.is_key:
                color = item.color
                colors_match = (entity.held_item.color == item.color)
                print color, colors_match
                if colors_match:
                    del self.entities[item.ID]
                    del self.entities[entity.held_item.ID]
                    entity.held_item = None
                    self.world[x][y].contents = None
                    self.Log("success","Dropped a key on the "+color+" door to open it.")
                    return True
            
            self.Log("fail","Target tile must be empty.")
            return False

        ## OK!
        dropped_item = entity.Drop()
        self.world[x][y].Put(dropped_item)
        self.Log("success","Dropped #"+dropped_item.ID+ " "+dropped_item.name+" to "+direction+".")
        return True

    def SetTestTerrain(self):
        self.world[4][2].terrain = "g"
        self.world[5][5].terrain = "g"
        self.world[7][1].terrain = "g"

    def Test(self):
        petunias = self.Create(name="Bowl of Petunias",coords=(4,2))
        assert petunias.coords == (4,2)

        assert self.world[4][2].contents is petunias
        assert self.world[4][3].contents == None
        self.Move(petunias,"e")
        assert petunias.coords == (5,2)
        self.Move(petunias,"s")
        assert petunias.coords == (5,3)

        whale_error = self.Create(name="Whale",coords=(5,3))
        if whale_error:
            raise SystemError("Cetacean Superposition Error.")
        whale = self.Create(name="Whale",coords=(5,4))
        assert self.world[5][4].contents is whale
        assert whale.coords == (5,4)
        assert self.Move(whale,"n") == False
        assert self.Move(whale,"w") == True
        self.Move(whale,"e")
        assert self.Take(whale,"n")
        assert self.Drop(whale,"s")
        assert self.world[5][5].contents
        print "Test complete. Log follows; errors are intentional."
        for line in self.log:
            print line

    def TextDisplay(self):
        """Text-based display of the game world."""
        text = "*"+("-"*self.size[0])+"*\n|"
        for y in range(self.size[1]):
            for x in range(self.size[0]):
                contents = self.world[x][y].contents
                if contents:
                    t = contents.name[0].upper()
                else:
                    t = self.world[x][y].terrain[0]
                text += t
            text += "|\n|"
        text = text[:-1]
        text += "*"+("-"*self.size[0])+"*"
        print text

    def Display(self):
        """Graphical display, in case no better one is used."""
        if not self.use_graphics:
            raise SystemError("You need to create the World object with option 'graphics'.")
        
        for x in range(self.size[0]):
            for y in range(self.size[1]):
                t = self.tiles[self.world[x][y].terrain]
                screen.blit(t,(x*50,y*50))
        pygame.display.update()

    def TakeCommand(self,sender,command):
        """Act on a command string given by a Character."""
##        self.Log("action","#"+str(sender.ID)+" "+sender.name+": "+command)
        self.Log("action,comic",sender.name+": "+command)
##        self.comic_script[-1].append(sender.name+": "+command)
        words = command.split(" ")
        params = len(words) - 1 ## For asking whether the right # of options was given.
        if not words:
            self.Log("errorSYN",100,"Got a blank command string!")
            return
        verb = words[0]

        ## Wait/Idle: Do nothing.
        if verb in ("wait","idle"):
            self.Log("success",sender.name+" waits.")
            self.AllSense(sender.ID,"idle ok")

        ## Go/Move: Take a step.
        elif verb in ("go","move"):
            if not params:
                self.Log("error,syntax","Syntax: 'go n/s/e/w'")
                return
            ok = self.Move(sender,words[1])
            if ok:
                self.AllSense(sender.ID,"move ok")
            else:
                self.AllSense(sender.ID,"move fail")

        ## Say/Talk: Say something.
        ## This should eventually pass messages to other Characters.
        elif verb in ("say","talk"):
            if not params:
                self.Log("error,syntax","Syntax: 'say [text]'")
                return
            text = " ".join(words[1:])
            self.Log("success,comic",sender.name+" says, "+text+"")
            sender.Sense(sender.ID,"say ok") ## Speaker gets success message
            self.AllSense(sender.ID,"says "+text) ## Everyone "hears"

        ## Take/Get: Pick up an adjacent object.
        elif verb in ("get","take"):
            if not params:
                self.Log("error,syntax","Syntax: 'get [dir]'")
            success = self.Take(sender,words[1])
            if success:
                sender.Sense(sender.ID,"get ok")
                self.AllSense(sender.ID,"gets "+sender.held_item.ID)
                self.Log("success,comic",sender.name+" gets "+sender.held_item.name+" from "+words[1]+".")
            else:
                self.AllSense(sender.ID,"get fail")

        ## Drop/Put: Drop currently held object.
        elif verb in ("drop","put"):
            if not params:
                self.Log("error,syntax","Syntax: 'drop [dir]'")
            item_name = None
            if sender.held_item:
                item = sender.held_item
            success = self.Drop(sender,words[1])
            if success:
                self.AllSense(sender.ID,"drops "+item.ID)
                self.Log("success,comic",sender.name+" drops "+item.ID+" to "+words[1]+".")
            else:
                self.AllSense(sender.ID,"drop fail")

        ## Look: Get sense info on an entity, referenced by ID#.
        elif verb == "look":
            success = False
            if not params:
                self.Log("error,syntax","Syntax: 'look [ID]'")
            ID = words[1].zfill(2) ## Make the ID two digits, eg. "01".
            if not ID in self.entities:
                self.Log("error","ID# is invalid.")
            else:
                success = True
                e = self.entities[ID]
                text = "see shape:"+e.shape+"/color:"+e.color+"/size:"+str(e.size)
                sender.Sense(e.ID,text)
            if success:
                self.AllSense(sender.ID,"looks_at "+ID)
                self.Log("success,comic",sender.name+" looks at "+self.entities[ID].ID+" "+self.entities[ID].name+".")
            else:
                self.AllSense(sender.ID,"look fail")

        ## Eat: Eat a held item.
        elif verb == "eat":
            success = False
            item = sender.held_item
            if not item:
                self.Log("error","Must be holding an item to eat it.")
            food_value = item.food_value
            if food_value > 0:
                success = True
                sender.ChangeEnergy(food_value)
            elif food_value < 0:
                success = True
                sender.ChangeEnergy(food_value)
                """If this kills the character, that'll be noticed when
                energy gets drained later in this function."""
            else:
                self.Log("error","The "+item.name+" is inedible.")
            if success:
                sender.held_item = None
                del self.entities[item.ID]
                self.Log("success,comic",sender.name+" eats "+item.name+". Food value: "+str(food_value))
                self.AllSense(sender.ID,sender.name+" eats "+item.ID)
            else:
                self.AllSense(sender.ID,"eat fail")

        ## Unrecognized verb.
        else:
            self.Log("error,syntax","Unrecognized command.")
            return

        ## Energy rule
        if self.rules["energy"]:
            print "Draining energy from "+sender.name
            drain = self.rules["energy drain rate"]
            survived = sender.ChangeEnergy(-drain)
            if not survived:
                self.Kill(sender)

    def AllSense(self,sender_ID,what):
        """Tell all characters about an event."""
        for c in self.characters:
            c.Sense(sender_ID,what)

    def SenseWorld(self,character):
        """Return data representing what can be seen.

        For now, assume that any character can see everything. That means
        unlimited vision range, plus awareness of entities held by others."""
        if character.vision_range:
            pass ## NYI
        else:
            ## Character can see everything
            entities = []
            text = "$ layout: "
##            for x in range(self.size[0]):
##                for y in range(self.size[1]):
##                    text += self.world[x][y].terrain[0]
##                    entity = self.world[x][y].contents
##                    if entity:
##                        e_text = "@ "+str(x)+","+str(y)+": "+str(entity.ID)+" "+entity.name
##                        if entity.alive:
##                            e_text += " Alive"
##                        entities.append(e_text)
            for y in range(self.size[1]):
                for x in range(self.size[0]):
                    text += self.world[x][y].terrain[0]
                    entity = self.world[x][y].contents
                    if entity:
                        e_text = "@ "+str(x)+","+str(y)+": "+str(entity.ID)+" "+entity.name
                        if entity.alive:
                            e_text += " Alive"
                        entities.append(e_text)
                ## Make the layout slightly human-readable while still being 1 line.
                text += "|" ## Strip | when reading this output.
            text += "\n"
            for entry in entities:
                text += entry+"\n"
            return text


##___________________________________
##            ~Functions~
##___________________________________


##__________________________________
##            ~Autorun~
##__________________________________

if __name__ == "__main__":
    print "Running a demo of some random terrain."
    print "Make sure GRAPHICS is set to True in the Constants section if you want to try the graphical demo."
    w = World(use_graphics=GRAPHICS)
    w.SetRandomTerrain()
    if GRAPHICS:
        w.Display()
    else:
        w.TextDisplay()
    w.WriteLog()

